package org.jaudiotagger.audio.wav.chunk;

import org.jaudiotagger.tag.FieldKey;

import java.util.HashMap;
import java.util.Map;

/**
 * Known Identifiers used in an INFO Chunk together with their mapping to a generic FieldKey (if known)
 */
public enum WavInfoIdentifier
{
    ARTIST("IART", FieldKey.ARTIST),
    ALBUM("IPRD", FieldKey.ALBUM),
    TITLE("INAM", FieldKey.TITLE),
    TRACKNO("ITRK", FieldKey.TRACK),
    YEAR("ICRD", FieldKey.YEAR),
    GENRE("IGNR", FieldKey.GENRE),
    ALBUM_ARTIST("iaar", FieldKey.ALBUM_ARTIST), //Dont know why lowercase, maybe MediaMonkey custom field
    COMMENTS("ICMT", FieldKey.COMMENT),
    COMPOSER("IMUS", FieldKey.COMPOSER),
    CONDUCTOR("ITCH", FieldKey.CONDUCTOR),
    LYRICIST("IWRI", FieldKey.LYRICIST),
    ENCODER("ISFT", FieldKey.ENCODER),
    RATING("IRTD", FieldKey.RATING),
    ISRC("ISRC", FieldKey.ISRC),
    LABEL("ICMS", FieldKey.RECORD_LABEL),
    TRACK_GAIN("ITGL", null), //Currently No mapping to a FieldKey for this
    ALBUM_GAIN("IAGL", null), //Currently No mapping to a FieldKey for this
    COPYRIGHT("ICOP", null),
    TWONKY_TRACKNO("itrk", null); //Uses nonstandard field

    private static final Map<String, WavInfoIdentifier> CODE_TYPE_MAP = new HashMap<String, WavInfoIdentifier>();
    private static final Map<FieldKey, WavInfoIdentifier> FIELDKEY_TYPE_MAP = new HashMap<FieldKey, WavInfoIdentifier>();
    private String code;
    private FieldKey fieldKey;

    WavInfoIdentifier(String code, FieldKey fieldKey)
    {
        this.code = code;
        this.fieldKey = fieldKey;
    }

    public String getCode()
    {
        return code;
    }

    public FieldKey getFieldKey()
    {
        return fieldKey;
    }


    /**
     * Get {@link WavInfoIdentifier} for code (e.g. "SSND").
     *
     * @param code chunk id
     * @return chunk type or {@code null} if not registered
     */
    public synchronized static WavInfoIdentifier getByCode(final String code)
    {
        if (CODE_TYPE_MAP.isEmpty())
        {
            for (final WavInfoIdentifier type : values())
            {
                CODE_TYPE_MAP.put(type.getCode(), type);
            }
        }
        return CODE_TYPE_MAP.get(code);
    }

    /**
     * Get {@link WavInfoIdentifier} for code (e.g. "SSND").
     *
     * @param fieldKey
     * @return chunk type or {@code null} if not registered
     */
    public synchronized static WavInfoIdentifier getByByFieldKey(final FieldKey fieldKey)
    {
        if (FIELDKEY_TYPE_MAP.isEmpty())
        {
            for (final WavInfoIdentifier type : values())
            {
                if (type.getFieldKey() != null)
                {
                    FIELDKEY_TYPE_MAP.put(type.getFieldKey(), type);
                }
            }
        }
        return FIELDKEY_TYPE_MAP.get(fieldKey);
    }
}
